-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE IEEE.NUMERIC_STD.ALL;     

PACKAGE ddr_pkg IS  
  
  -- DDR3 and DDR4 widths
  TYPE t_c_ddr_phy IS RECORD
    a_w                               : NATURAL;  -- = 16;
    a_row_w                           : NATURAL;  -- = 16;  --  = a_w, row address width, via a_w lines
    a_col_w                           : NATURAL;  -- = 10;  -- <= a_w, col address width, via a_w lines
    ba_w                              : NATURAL;  -- = 2;
    bg_w                              : NATURAL;  -- = 2;
    dq_w                              : NATURAL;  -- = 64;
    dqs_w                             : NATURAL;  -- = 8;   --  = dq_w / nof_dq_per_dqs;
    dm_w                              : NATURAL;  -- = 8;
    cs_w                              : NATURAL;  -- = 2;
    act_w                             : NATURAL;  -- = 1;
    odt_w                             : NATURAL;  -- = 2;
    resetarr_w                        : NATURAL;  -- = 1;
    alert_w                           : NATURAL;  -- = 1;
    par_w                             : NATURAL;  -- = 1;
    cb_w                              : NATURAL;  -- = 8;
  END RECORD;

  CONSTANT c_ddr4_phy_all             : t_c_ddr_phy := (17, 0, 0, 2, 2, 64, 9, 9, 2, 1, 2, 1, 1, 1, 8);
  CONSTANT c_ddr4_phy_4g              : t_c_ddr_phy := (17, 0, 0, 2, 2, 64, 9, 9, 1, 1, 1, 1, 1, 1, 8);

  TYPE t_ddr_phy_in IS RECORD  
    evt              : STD_LOGIC;
    alert_n          : STD_LOGIC_VECTOR(c_ddr_phy.alert_w-1 DOWNTO 0);  
    oct_rup          : STD_LOGIC;
    oct_rdn          : STD_LOGIC;
    oct_rzqin        : STD_LOGIC;
    nc               : STD_LOGIC;   -- not connected, needed to be able to initialize constant record which has to have more than one field in VHDL
  END RECORD;
  
  TYPE t_ddr_phy_io IS RECORD -- Do not use this type in Quartus! Use the _sel version instead.
    dq               : STD_LOGIC_VECTOR(c_ddr_phy.dq_w-1 DOWNTO 0);   -- data bus
    cb               : STD_LOGIC_VECTOR(c_ddr_phy.cb_w-1 DOWNTO 0);   -- data bus
    dqs              : STD_LOGIC_VECTOR(c_ddr_phy.dqs_w-1 DOWNTO 0);  -- data strobe bus
    dqs_n            : STD_LOGIC_VECTOR(c_ddr_phy.dqs_w-1 DOWNTO 0);
    clk              : STD_LOGIC_VECTOR(c_ddr_phy.clk_w-1 DOWNTO 0);  -- clock, positive edge clock
    clk_n            : STD_LOGIC_VECTOR(c_ddr_phy.clk_w-1 DOWNTO 0);  -- clock, negative edge clock
    scl              : STD_LOGIC;                                      -- I2C
    sda              : STD_LOGIC;
  END RECORD;
  
  TYPE t_ddr_phy_ou IS RECORD  
    a                : STD_LOGIC_VECTOR(c_ddr_phy.a_w-1 DOWNTO 0);    -- row and column address
    ba               : STD_LOGIC_VECTOR(c_ddr_phy.ba_w-1 DOWNTO 0);   -- bank address
    bg               : STD_LOGIC_VECTOR(c_ddr_phy.bg_w-1 DOWNTO 0);   -- bank 
    dm               : STD_LOGIC_VECTOR(c_ddr_phy.dm_w-1 DOWNTO 0);   -- data mask bus
    cas_n            : STD_LOGIC; --_VECTOR(0 DOWNTO 0);                   -- column address strobe
    ras_n            : STD_LOGIC; --_VECTOR(0 DOWNTO 0);                   -- row address strobe
    we_n             : STD_LOGIC; --_VECTOR(0 DOWNTO 0);                   -- write enable signal
    reset_n          : STD_LOGIC;                                      -- reset signal
    resetarr_n       : STD_LOGIC_VECTOR(c_ddr_phy.resetarr_w-1 DOWNTO 0);   -- on-die termination control signal
    par              : STD_LOGIC_VECTOR(c_ddr_phy.par_w-1 DOWNTO 0);   -- on-die termination control signal
    odt              : STD_LOGIC_VECTOR(c_ddr_phy.cs_w-1 DOWNTO 0);   -- on-die termination control signal
    cke              : STD_LOGIC_VECTOR(c_ddr_phy.cs_w-1 DOWNTO 0);   -- clock enable
    cs_n             : STD_LOGIC_VECTOR(c_ddr_phy.cs_w-1 DOWNTO 0);   -- chip select
  END RECORD;

  CONSTANT c_ddr_phy_in_rst   : t_ddr_phy_in := ('0', '1', 'X', 'X', 'X', 'X');
  CONSTANT c_ddr_phy_io_rst   : t_ddr_phy_io := ((OTHERS=>'0'), (others => '0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), '0', '0');
  CONSTANT c_ddr_phy_ou_rst   : t_ddr_phy_ou := ((OTHERS=>'0'), (OTHERS=>'0'), (others => '0'), (OTHERS=>'0'), '0', '0', '0', '0', (others => '0'), (others => '0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'));

  TYPE t_ddr_phy_in_arr IS ARRAY(NATURAL RANGE <>) OF t_ddr_phy_in;
  TYPE t_ddr_phy_io_arr IS ARRAY(NATURAL RANGE <>) OF t_ddr_phy_io; 
  TYPE t_ddr_phy_ou_arr IS ARRAY(NATURAL RANGE <>) OF t_ddr_phy_ou;

  TYPE t_ddr_addr IS RECORD 
    chip      : STD_LOGIC_VECTOR(ceil_log2(c_ddr_phy.cs_w)  -1 DOWNTO 0);  -- Note: The controller interprets the chip address as logical address (NOT individual chip sel lines), hence ceil_log2
    bank      : STD_LOGIC_VECTOR(          c_ddr_phy.ba_w   -1 DOWNTO 0);
    row       : STD_LOGIC_VECTOR(          c_ddr_phy.a_row_w-1 DOWNTO 0); 
    column    : STD_LOGIC_VECTOR(          c_ddr_phy.a_col_w-1 DOWNTO 0);
  END RECORD;

  TYPE t_ddr_addr_arr IS ARRAY(NATURAL RANGE <>) OF t_ddr_addr; 

  constant c_ddr_local_data_w : integer := sel_n(g_phy, 256, 256, 256, 576)
  --CONSTANT c_ddr_ctlr_data_w            : NATURAL := 256;  -- = 64 (PHY dq width) * 2 (use both PHY clock edges) * 2 (PHY transfer at double rate) -- 512 for DDR4
  CONSTANT c_ddr_ctlr_rsl              : NATURAL := c_ddr_local_data_w / (c_ddr_phy.dq_w + c_ddr_phy.cb_w); -- =4 or 8
  CONSTANT c_ddr_ctlr_rsl_w            : NATURAL := ceil_log2(c_ddr_ctlr_rsl);  -- 2 or 3
  CONSTANT c_ddr_ctlr_maxburstsize     : NATURAL := 64;
  CONSTANT c_ddr_ctlr_maxburstsize_w   : NATURAL := ceil_log2(c_ddr_ctlr_maxburstsize+1);
  CONSTANT c_ddr_ctrl_nof_latent_reads : NATURAL := 100;  -- The downside to having a command cue: even after de-asserting read requests, the ALTMEMPHY keeps processing your cued read requests. 
                                                           -- This makes sure 100 words are still available in the read FIFO after it de-asserted its siso.ready signal towards the ddr3 read side.


  CONSTANT c_ddr_addr_lo               : t_ddr_addr := ((OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'),  (OTHERS=>'0'));
  CONSTANT c_ddr_address_lo            : NATURAL := 0;

  CONSTANT c_ddr_addr_hi_4gb           : t_ddr_addr := ((OTHERS=>'1'), (OTHERS=>'1'), (OTHERS=>'1'),  TO_UVEC(2**c_ddr_phy_4g.a_col_w - c_ddr_ctlr_rsl, c_ddr_phy_4g.a_col_w));

  -- 4 rows * 1024 cols  * 64 bits / 128bits per associative mem array element = 2048 = default ALTMEMPHY mem_model array depth.
  CONSTANT c_ddr_addr_hi_sim           : t_ddr_addr := ((OTHERS=>'0'), (OTHERS=>'0'), TO_UVEC(3, c_ddr_phy.a_row_w),  TO_UVEC(2**c_ddr_phy_4g.a_col_w - c_ddr_ctlr_rsl, c_ddr_phy_4g.a_col_w));
  CONSTANT c_ddr_address_hi_sim        : NATURAL := 4092; --TB uses generated mem model with 2ki addresses of 128k - so the array holds 4096 64-bit words. End address is 4092 (resolution=4: last write=4092,4093,4094,4095.)
  
  TYPE t_ddr_seq IS RECORD 
    g_wr_chunksize    : POSITIVE; -- := 256;
    g_wr_nof_steps    : POSITIVE; -- := 1;
    g_wr_nof_blocks   : POSITIVE; -- := 64;
    g_rd_chunksize    : POSITIVE; -- := 8;
    g_rd_nof_steps    : POSITIVE; -- := 16;
    g_rd_nof_blocks   : POSITIVE; -- := 128;   
    g_switch_size     : POSITIVE; -- := 256;    -- This defines the switch interval between read and write operation. 
    g_page_size       : POSITIVE; -- := 16384;  -- Page size defines the size of a single page. Total memory will be doubled. 
  END RECORD;

  CONSTANT c_ddr_seq : t_ddr_seq := (256, 1, 64, 8, 16, 128, 256, 16384);  
  
  
  COMPONENT uphy_4g_800_master IS
	PORT (
		pll_ref_clk                : IN    STD_LOGIC := '0';               -- pll_ref_clk.clk
		global_reset_n             : IN    STD_LOGIC := '0';               -- global_reset.reset_n
		soft_reset_n               : IN    STD_LOGIC := '0';               -- soft_reset.reset_n
		afi_clk                    : OUT   STD_LOGIC;                      -- afi_clk.clk
		afi_half_clk               : OUT   STD_LOGIC;                      -- afi_half_clk.clk
		afi_reset_n                : OUT   STD_LOGIC;                      -- afi_reset.reset_n
		mem_a                      : OUT   STD_LOGIC_VECTOR(14 DOWNTO 0);  -- memory.mem_a
		mem_ba                     : OUT   STD_LOGIC_VECTOR(2 DOWNTO 0);   -- .mem_ba
		mem_ck                     : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_ck
		mem_ck_n                   : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_ck_n
		mem_cke                    : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_cke
		mem_cs_n                   : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_cs_n
		mem_dm                     : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0);   -- .mem_dm
		mem_ras_n                  : OUT   STD_LOGIC;                      -- .mem_ras_n
		mem_cas_n                  : OUT   STD_LOGIC;                      -- .mem_cas_n
		mem_we_n                   : OUT   STD_LOGIC;                      -- .mem_we_n
		mem_reset_n                : OUT   STD_LOGIC;                      -- .mem_reset_n
		mem_dq                     : INOUT STD_LOGIC_VECTOR(63 DOWNTO 0) := (OTHERS => '0'); -- .mem_dq
		mem_dqs                    : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => '0'); -- .mem_dqs
		mem_dqs_n                  : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => '0'); -- .mem_dqs_n
		mem_odt                    : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);                     -- .mem_odt
		avl_ready                  : OUT   STD_LOGIC;                                        -- avl.waitrequest_n
		avl_burstbegin             : IN    STD_LOGIC  := '0';                                -- .beginbursttransfer
		avl_addr                   : IN    STD_LOGIC_VECTOR(26 DOWNTO 0) := (OTHERS => '0'); -- .address
		avl_rdata_valid            : OUT   STD_LOGIC;                                        -- .readdatavalid
		avl_rdata                  : OUT   STD_LOGIC_VECTOR(255 DOWNTO 0);                   -- .readdata
		avl_wdata                  : IN    STD_LOGIC_VECTOR(255 DOWNTO 0):= (OTHERS => '0'); -- .writedata
		avl_be                     : IN    STD_LOGIC_VECTOR(31 DOWNTO 0) := (OTHERS => '0'); -- .byteenable
		avl_read_req               : IN    STD_LOGIC := '0';                                 -- .read
		avl_write_req              : IN    STD_LOGIC := '0';                                 -- .write
		avl_size                   : IN    STD_LOGIC_VECTOR(6 DOWNTO 0)  := (OTHERS => '0'); -- .burstcount
		local_init_done            : OUT   STD_LOGIC;                                        -- status.local_init_done
		local_cal_success          : OUT   STD_LOGIC;                                        -- .local_cal_success
		local_cal_fail             : OUT   STD_LOGIC;                                        -- .local_cal_fail
		oct_rdn                    : IN    STD_LOGIC := '0';                                 -- oct.rdn
		oct_rup                    : IN    STD_LOGIC := '0';                                 -- oct.rup
		seriesterminationcontrol   : OUT   STD_LOGIC_VECTOR(13 DOWNTO 0);                    -- oct_sharing.seriesterminationcontrol
		parallelterminationcontrol : OUT   STD_LOGIC_VECTOR(13 DOWNTO 0);                    --            .parallelterminationcontrol
  	pll_mem_clk                : OUT   STD_LOGIC;                     -- export
		pll_write_clk              : OUT   STD_LOGIC;                     -- export
		pll_write_clk_pre_phy_clk  : OUT   STD_LOGIC;                     -- export
		pll_addr_cmd_clk           : OUT   STD_LOGIC;                     -- export
		pll_locked                 : OUT   STD_LOGIC;                     -- export
		pll_avl_clk                : OUT   STD_LOGIC;                     -- export
		pll_config_clk             : OUT   STD_LOGIC;                     -- export
		dll_delayctrl              : OUT   STD_LOGIC_VECTOR(5 DOWNTO 0)   -- export
	);
  END COMPONENT uphy_4g_800_master;

  COMPONENT uphy_4g_800_slave IS
	PORT (
		pll_ref_clk                : IN    STD_LOGIC := '0';               -- pll_ref_clk.clk
		global_reset_n             : IN    STD_LOGIC := '0';               -- global_reset.reset_n
		soft_reset_n               : IN    STD_LOGIC := '0';               -- soft_reset.reset_n
		afi_clk                    : OUT   STD_LOGIC;                      -- afi_clk.clk
		afi_half_clk               : OUT   STD_LOGIC;                      -- afi_half_clk.clk
		afi_reset_n                : OUT   STD_LOGIC;                      -- afi_reset.reset_n
		mem_a                      : OUT   STD_LOGIC_VECTOR(14 DOWNTO 0);  -- memory.mem_a
		mem_ba                     : OUT   STD_LOGIC_VECTOR(2 DOWNTO 0);   -- .mem_ba
		mem_ck                     : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_ck
		mem_ck_n                   : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_ck_n
		mem_cke                    : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_cke
		mem_cs_n                   : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_cs_n
		mem_dm                     : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0);   -- .mem_dm
		mem_ras_n                  : OUT   STD_LOGIC;                      -- .mem_ras_n
		mem_cas_n                  : OUT   STD_LOGIC;                      -- .mem_cas_n
		mem_we_n                   : OUT   STD_LOGIC;                      -- .mem_we_n
		mem_reset_n                : OUT   STD_LOGIC;                      -- .mem_reset_n
		mem_dq                     : INOUT STD_LOGIC_VECTOR(63 DOWNTO 0) := (OTHERS => '0'); -- .mem_dq
		mem_dqs                    : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => '0'); -- .mem_dqs
		mem_dqs_n                  : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => '0'); -- .mem_dqs_n
		mem_odt                    : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);                     -- .mem_odt
		avl_ready                  : OUT   STD_LOGIC;                                        -- avl.waitrequest_n
		avl_burstbegin             : IN    STD_LOGIC := '0';                                 -- .beginbursttransfer
		avl_addr                   : IN    STD_LOGIC_VECTOR(26 DOWNTO 0) := (OTHERS => '0'); -- .address
		avl_rdata_valid            : OUT   STD_LOGIC;                                        -- .readdatavalid
		avl_rdata                  : OUT   STD_LOGIC_VECTOR(255 DOWNTO 0);                   -- .readdata
		avl_wdata                  : IN    STD_LOGIC_VECTOR(255 DOWNTO 0):= (OTHERS => '0'); -- .writedata
		avl_be                     : IN    STD_LOGIC_VECTOR(31 DOWNTO 0) := (OTHERS => '0'); -- .byteenable
		avl_read_req               : IN    STD_LOGIC := '0';                                 -- .read
		avl_write_req              : IN    STD_LOGIC := '0';                                 -- .write
		avl_size                   : IN    STD_LOGIC_VECTOR(6 DOWNTO 0)  := (OTHERS => '0'); -- .burstcount
		local_init_done            : OUT   STD_LOGIC;                                        -- status.local_init_done
		local_cal_success          : OUT   STD_LOGIC;                                        -- .local_cal_success
		local_cal_fail             : OUT   STD_LOGIC;                                        -- .local_cal_fail
		seriesterminationcontrol   : IN    STD_LOGIC_VECTOR(13 DOWNTO 0);                    -- oct_sharing.seriesterminationcontrol
		parallelterminationcontrol : IN    STD_LOGIC_VECTOR(13 DOWNTO 0);                    --            .parallelterminationcontrol
  	pll_mem_clk                : OUT   STD_LOGIC;                     -- export
		pll_write_clk              : OUT   STD_LOGIC;                     -- export
		pll_write_clk_pre_phy_clk  : OUT   STD_LOGIC;                     -- export
		pll_addr_cmd_clk           : OUT   STD_LOGIC;                     -- export
		pll_locked                 : OUT   STD_LOGIC;                     -- export
		pll_avl_clk                : OUT   STD_LOGIC;                     -- export
		pll_config_clk             : OUT   STD_LOGIC;                     -- export
		dll_delayctrl              : OUT   STD_LOGIC_VECTOR(5 DOWNTO 0)   -- export
	);
  END COMPONENT uphy_4g_800_slave;

  COMPONENT uphy_4g_1066_master IS
	PORT (
		pll_ref_clk                : IN    STD_LOGIC := '0';               -- pll_ref_clk.clk
		global_reset_n             : IN    STD_LOGIC := '0';               -- global_reset.reset_n
		soft_reset_n               : IN    STD_LOGIC := '0';               -- soft_reset.reset_n
		afi_clk                    : OUT   STD_LOGIC;                      -- afi_clk.clk
		afi_half_clk               : OUT   STD_LOGIC;                      -- afi_half_clk.clk
		afi_reset_n                : OUT   STD_LOGIC;                      -- afi_reset.reset_n
		mem_a                      : OUT   STD_LOGIC_VECTOR(14 DOWNTO 0);  -- memory.mem_a
		mem_ba                     : OUT   STD_LOGIC_VECTOR(2 DOWNTO 0);   -- .mem_ba
		mem_ck                     : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_ck
		mem_ck_n                   : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_ck_n
		mem_cke                    : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_cke
		mem_cs_n                   : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_cs_n
		mem_dm                     : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0);   -- .mem_dm
		mem_ras_n                  : OUT   STD_LOGIC;                      -- .mem_ras_n
		mem_cas_n                  : OUT   STD_LOGIC;                      -- .mem_cas_n
		mem_we_n                   : OUT   STD_LOGIC;                      -- .mem_we_n
		mem_reset_n                : OUT   STD_LOGIC;                      -- .mem_reset_n
		mem_dq                     : INOUT STD_LOGIC_VECTOR(63 DOWNTO 0) := (OTHERS => '0'); -- .mem_dq
		mem_dqs                    : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => '0'); -- .mem_dqs
		mem_dqs_n                  : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => '0'); -- .mem_dqs_n
		mem_odt                    : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);                     -- .mem_odt
		avl_ready                  : OUT   STD_LOGIC;                                        -- avl.waitrequest_n
		avl_burstbegin             : IN    STD_LOGIC  := '0';                                -- .beginbursttransfer
		avl_addr                   : IN    STD_LOGIC_VECTOR(26 DOWNTO 0) := (OTHERS => '0'); -- .address
		avl_rdata_valid            : OUT   STD_LOGIC;                                        -- .readdatavalid
		avl_rdata                  : OUT   STD_LOGIC_VECTOR(255 DOWNTO 0);                   -- .readdata
		avl_wdata                  : IN    STD_LOGIC_VECTOR(255 DOWNTO 0):= (OTHERS => '0'); -- .writedata
		avl_be                     : IN    STD_LOGIC_VECTOR(31 DOWNTO 0) := (OTHERS => '0'); -- .byteenable
		avl_read_req               : IN    STD_LOGIC := '0';                                 -- .read
		avl_write_req              : IN    STD_LOGIC := '0';                                 -- .write
		avl_size                   : IN    STD_LOGIC_VECTOR(6 DOWNTO 0)  := (OTHERS => '0'); -- .burstcount
		local_init_done            : OUT   STD_LOGIC;                                        -- status.local_init_done
		local_cal_success          : OUT   STD_LOGIC;                                        -- .local_cal_success
		local_cal_fail             : OUT   STD_LOGIC;                                        -- .local_cal_fail
		oct_rdn                    : IN    STD_LOGIC := '0';                                 -- oct.rdn
		oct_rup                    : IN    STD_LOGIC := '0';                                 -- oct.rup
		seriesterminationcontrol   : OUT   STD_LOGIC_VECTOR(13 DOWNTO 0);                    -- oct_sharing.seriesterminationcontrol
		parallelterminationcontrol : OUT   STD_LOGIC_VECTOR(13 DOWNTO 0);                    --            .parallelterminationcontrol
  	pll_mem_clk                : OUT   STD_LOGIC;                     -- export
		pll_write_clk              : OUT   STD_LOGIC;                     -- export
		pll_write_clk_pre_phy_clk  : OUT   STD_LOGIC;                     -- export
		pll_addr_cmd_clk           : OUT   STD_LOGIC;                     -- export
		pll_locked                 : OUT   STD_LOGIC;                     -- export
		pll_avl_clk                : OUT   STD_LOGIC;                     -- export
		pll_config_clk             : OUT   STD_LOGIC;                     -- export
		dll_delayctrl              : OUT   STD_LOGIC_VECTOR(5 DOWNTO 0)   -- export
	);
  END COMPONENT uphy_4g_1066_master;
  
  COMPONENT uphy_4g_1066_slave IS
	PORT (
		pll_ref_clk                : IN    STD_LOGIC := '0';               -- pll_ref_clk.clk
		global_reset_n             : IN    STD_LOGIC := '0';               -- global_reset.reset_n
		soft_reset_n               : IN    STD_LOGIC := '0';               -- soft_reset.reset_n
		afi_clk                    : OUT   STD_LOGIC;                      -- afi_clk.clk
		afi_half_clk               : OUT   STD_LOGIC;                      -- afi_half_clk.clk
		afi_reset_n                : OUT   STD_LOGIC;                      -- afi_reset.reset_n
		mem_a                      : OUT   STD_LOGIC_VECTOR(14 DOWNTO 0);  -- memory.mem_a
		mem_ba                     : OUT   STD_LOGIC_VECTOR(2 DOWNTO 0);   -- .mem_ba
		mem_ck                     : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_ck
		mem_ck_n                   : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_ck_n
		mem_cke                    : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_cke
		mem_cs_n                   : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);   -- .mem_cs_n
		mem_dm                     : OUT   STD_LOGIC_VECTOR(7 DOWNTO 0);   -- .mem_dm
		mem_ras_n                  : OUT   STD_LOGIC;                      -- .mem_ras_n
		mem_cas_n                  : OUT   STD_LOGIC;                      -- .mem_cas_n
		mem_we_n                   : OUT   STD_LOGIC;                      -- .mem_we_n
		mem_reset_n                : OUT   STD_LOGIC;                      -- .mem_reset_n
		mem_dq                     : INOUT STD_LOGIC_VECTOR(63 DOWNTO 0) := (OTHERS => '0'); -- .mem_dq
		mem_dqs                    : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => '0'); -- .mem_dqs
		mem_dqs_n                  : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => '0'); -- .mem_dqs_n
		mem_odt                    : OUT   STD_LOGIC_VECTOR(1 DOWNTO 0);                     -- .mem_odt
		avl_ready                  : OUT   STD_LOGIC;                                        -- avl.waitrequest_n
		avl_burstbegin             : IN    STD_LOGIC := '0';                                 -- .beginbursttransfer
		avl_addr                   : IN    STD_LOGIC_VECTOR(26 DOWNTO 0) := (OTHERS => '0'); -- .address
		avl_rdata_valid            : OUT   STD_LOGIC;                                        -- .readdatavalid
		avl_rdata                  : OUT   STD_LOGIC_VECTOR(255 DOWNTO 0);                   -- .readdata
		avl_wdata                  : IN    STD_LOGIC_VECTOR(255 DOWNTO 0):= (OTHERS => '0'); -- .writedata
		avl_be                     : IN    STD_LOGIC_VECTOR(31 DOWNTO 0) := (OTHERS => '0'); -- .byteenable
		avl_read_req               : IN    STD_LOGIC := '0';                                 -- .read
		avl_write_req              : IN    STD_LOGIC := '0';                                 -- .write
		avl_size                   : IN    STD_LOGIC_VECTOR(6 DOWNTO 0)  := (OTHERS => '0'); -- .burstcount
		local_init_done            : OUT   STD_LOGIC;                                        -- status.local_init_done
		local_cal_success          : OUT   STD_LOGIC;                                        -- .local_cal_success
		local_cal_fail             : OUT   STD_LOGIC;                                        -- .local_cal_fail
		seriesterminationcontrol   : IN    STD_LOGIC_VECTOR(13 DOWNTO 0);                    -- oct_sharing.seriesterminationcontrol
		parallelterminationcontrol : IN    STD_LOGIC_VECTOR(13 DOWNTO 0);                    --            .parallelterminationcontrol
  	pll_mem_clk                : OUT   STD_LOGIC;                     -- export
		pll_write_clk              : OUT   STD_LOGIC;                     -- export
		pll_write_clk_pre_phy_clk  : OUT   STD_LOGIC;                     -- export
		pll_addr_cmd_clk           : OUT   STD_LOGIC;                     -- export
		pll_locked                 : OUT   STD_LOGIC;                     -- export
		pll_avl_clk                : OUT   STD_LOGIC;                     -- export
		pll_config_clk             : OUT   STD_LOGIC;                     -- export
		dll_delayctrl              : OUT   STD_LOGIC_VECTOR(5 DOWNTO 0)   -- export
	);
  END COMPONENT uphy_4g_1066_slave;
  
  COMPONENT alt_mem_if_ddr3_mem_model_top_ddr3_mem_if_dm_pins_en_mem_if_dqsn_en IS
	GENERIC (
		MEM_IF_ADDR_WIDTH            : INTEGER := 0;
		MEM_IF_ROW_ADDR_WIDTH        : INTEGER := 0;
		MEM_IF_COL_ADDR_WIDTH        : INTEGER := 0;
		MEM_IF_CS_PER_RANK           : INTEGER := 0;
		MEM_IF_CONTROL_WIDTH         : INTEGER := 0;
		MEM_IF_DQS_WIDTH             : INTEGER := 0;
		MEM_IF_CS_WIDTH              : INTEGER := 0;
		MEM_IF_BANKADDR_WIDTH        : INTEGER := 0;
		MEM_IF_DQ_WIDTH              : INTEGER := 0;
		MEM_IF_CK_WIDTH              : INTEGER := 0;
		MEM_IF_CLK_EN_WIDTH          : INTEGER := 0;
		DEVICE_WIDTH                 : INTEGER := 1;
		MEM_TRCD                     : INTEGER := 0;
		MEM_TRTP                     : INTEGER := 0;
		MEM_DQS_TO_CLK_CAPTURE_DELAY : INTEGER := 0;
		MEM_CLK_TO_DQS_CAPTURE_DELAY : INTEGER := 0;
		MEM_IF_ODT_WIDTH             : INTEGER := 0;
		MEM_MIRROR_ADDRESSING_DEC    : INTEGER := 0;
		MEM_REGDIMM_ENABLED          : BOOLEAN := FALSE;
		DEVICE_DEPTH                 : INTEGER := 1;
		MEM_GUARANTEED_WRITE_INIT    : BOOLEAN := FALSE;
		MEM_VERBOSE                  : BOOLEAN := TRUE;
		MEM_INIT_EN                  : BOOLEAN := FALSE;
		MEM_INIT_FILE                : STRING  := "";
		DAT_DATA_WIDTH               : INTEGER := 32
	);
	PORT (
		mem_a       : IN    STD_LOGIC_VECTOR(14 DOWNTO 0) := (OTHERS => 'X'); -- mem_a
		mem_ba      : IN    STD_LOGIC_VECTOR(2 DOWNTO 0)  := (OTHERS => 'X'); -- mem_ba
		mem_ck      : IN    STD_LOGIC_VECTOR(1 DOWNTO 0)  := (OTHERS => 'X'); -- mem_ck
		mem_ck_n    : IN    STD_LOGIC_VECTOR(1 DOWNTO 0)  := (OTHERS => 'X'); -- mem_ck_n
		mem_cke     : IN    STD_LOGIC_VECTOR(1 DOWNTO 0)  := (OTHERS => 'X'); -- mem_cke
		mem_cs_n    : IN    STD_LOGIC_VECTOR(1 DOWNTO 0)  := (OTHERS => 'X'); -- mem_cs_n
		mem_dm      : IN    STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => 'X'); -- mem_dm
		mem_ras_n   : IN    STD_LOGIC_VECTOR(0 DOWNTO 0)  := (OTHERS => 'X'); -- mem_ras_n
		mem_cas_n   : IN    STD_LOGIC_VECTOR(0 DOWNTO 0)  := (OTHERS => 'X'); -- mem_cas_n
		mem_we_n    : IN    STD_LOGIC_VECTOR(0 DOWNTO 0)  := (OTHERS => 'X'); -- mem_we_n
		mem_reset_n : IN    STD_LOGIC                     := 'X';             -- mem_reset_n
		mem_dq      : INOUT STD_LOGIC_VECTOR(63 DOWNTO 0) := (OTHERS => 'X'); -- mem_dq
		mem_dqs     : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => 'X'); -- mem_dqs
		mem_dqs_n   : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0)  := (OTHERS => 'X'); -- mem_dqs_n
		mem_odt     : IN    STD_LOGIC_VECTOR(1 DOWNTO 0)  := (OTHERS => 'X')  -- mem_odt
	);
	END COMPONENT alt_mem_if_ddr3_mem_model_top_ddr3_mem_if_dm_pins_en_mem_if_dqsn_en;

  -- generic component using all possible IOs for pinning check
  component ddr4_all is
  port (
    global_reset_n      : in    std_logic                      := 'X';             -- reset_n
    pll_ref_clk         : in    std_logic                      := 'X';             -- clk
    oct_rzqin           : in    std_logic                      := 'X';             -- oct_rzqin
    mem_ck              : out   std_logic_vector(1 downto 0);                      -- mem_ck
    mem_ck_n            : out   std_logic_vector(1 downto 0);                      -- mem_ck_n
    mem_a               : out   std_logic_vector(16 downto 0);                     -- mem_a
    mem_act_n           : out   std_logic_vector(0 downto 0);                      -- mem_act_n
    mem_ba              : out   std_logic_vector(1 downto 0);                      -- mem_ba
    mem_bg              : out   std_logic_vector(1 downto 0);                      -- mem_bg
    mem_cke             : out   std_logic_vector(1 downto 0);                      -- mem_cke
    mem_cs_n            : out   std_logic_vector(1 downto 0);                      -- mem_cs_n
    mem_odt             : out   std_logic_vector(1 downto 0);                      -- mem_odt
    mem_reset_n         : out   std_logic_vector(0 downto 0);                      -- mem_reset_n
    mem_alert_n         : in   std_logic_vector(0 downto 0);                       -- mem_alert_n
    mem_par             : out   std_logic_vector(0 downto 0);                      -- mem_par  ** new in 14.0 **
    mem_dqs             : inout std_logic_vector(8 downto 0)   := (others => 'X'); -- mem_dqs
    mem_dqs_n           : inout std_logic_vector(8 downto 0)   := (others => 'X'); -- mem_dqs_n
    mem_dq              : inout std_logic_vector(71 downto 0)  := (others => 'X'); -- mem_dq
    mem_dbi_n           : inout std_logic_vector(8 downto 0)   := (others => 'X'); -- mem_dbi_n
    local_cal_success   : out   std_logic;                                         -- local_cal_success
    local_cal_fail      : out   std_logic;                                         -- local_cal_fail
    emif_usr_reset_n    : out   std_logic;                                         -- reset_n
    emif_usr_clk        : out   std_logic;                                         -- clk
    amm_ready_0         : out   std_logic;                                         -- waitrequest_n
    amm_read_0          : in    std_logic                      := 'X';             -- read
    amm_write_0         : in    std_logic                      := 'X';             -- write
    amm_address_0       : in    std_logic_vector(26 downto 0)  := (others => 'X'); -- address ** chg from 23 bits in 14.0 **
    amm_readdata_0      : out   std_logic_vector(575 downto 0);                    -- readdata
    amm_writedata_0     : in    std_logic_vector(575 downto 0) := (others => 'X'); -- writedata
    amm_burstcount_0    : in    std_logic_vector(6 downto 0)   := (others => 'X'); -- burstcount ** chg from 8 bits in 14.0 **
    amm_byteenable_0    : in    std_logic_vector(71 downto 0)  := (others => 'X'); -- byteenable
    amm_readdatavalid_0 : out   std_logic                                          -- readdatavalid
  );
  end component ddr4_all;

  -- Micron MTA9ASF51272PZ - 4GB
  -- clock speed 800MHz for testing
  component ddr4_4g_1600 is
  port (
    global_reset_n      : in    std_logic                      := 'X';             -- reset_n
    pll_ref_clk         : in    std_logic                      := 'X';             -- clk
    oct_rzqin           : in    std_logic                      := 'X';             -- oct_rzqin
    mem_ck              : out   std_logic_vector(0 downto 0);                      -- mem_ck
    mem_ck_n            : out   std_logic_vector(0 downto 0);                      -- mem_ck_n
    mem_a               : out   std_logic_vector(16 downto 0);                     -- mem_a
    mem_act_n           : out   std_logic_vector(0 downto 0);                      -- mem_act_n
    mem_ba              : out   std_logic_vector(1 downto 0);                      -- mem_ba
    mem_bg              : out   std_logic_vector(1 downto 0);                      -- mem_bg
    mem_cke             : out   std_logic_vector(0 downto 0);                      -- mem_cke
    mem_cs_n            : out   std_logic_vector(0 downto 0);                      -- mem_cs_n
    mem_odt             : out   std_logic_vector(0 downto 0);                      -- mem_odt
    mem_reset_n         : out   std_logic_vector(0 downto 0);                      -- mem_reset_n
    mem_alert_n         : in   std_logic_vector(0 downto 0);                       -- mem_alert_n
    mem_par             : out   std_logic_vector(0 downto 0);                      -- mem_par  ** new in 14.0 **
    mem_dqs             : inout std_logic_vector(8 downto 0)   := (others => 'X'); -- mem_dqs
    mem_dqs_n           : inout std_logic_vector(8 downto 0)   := (others => 'X'); -- mem_dqs_n
    mem_dq              : inout std_logic_vector(71 downto 0)  := (others => 'X'); -- mem_dq
    mem_dbi_n           : inout std_logic_vector(8 downto 0)   := (others => 'X'); -- mem_dbi_n
    local_cal_success   : out   std_logic;                                         -- local_cal_success
    local_cal_fail      : out   std_logic;                                         -- local_cal_fail
    emif_usr_reset_n    : out   std_logic;                                         -- reset_n
    emif_usr_clk        : out   std_logic;                                         -- clk
    amm_ready_0         : out   std_logic;                                         -- waitrequest_n
    amm_read_0          : in    std_logic                      := 'X';             -- read
    amm_write_0         : in    std_logic                      := 'X';             -- write
    amm_address_0       : in    std_logic_vector(25 downto 0)  := (others => 'X'); -- address ** chg from 23 bits in 14.0 **
    amm_readdata_0      : out   std_logic_vector(575 downto 0);                    -- readdata
    amm_writedata_0     : in    std_logic_vector(575 downto 0) := (others => 'X'); -- writedata
    amm_burstcount_0    : in    std_logic_vector(6 downto 0)   := (others => 'X'); -- burstcount ** chg from 8 bits in 14.0 **
    amm_byteenable_0    : in    std_logic_vector(71 downto 0)  := (others => 'X'); -- byteenable
    amm_readdatavalid_0 : out   std_logic                                          -- readdatavalid
  );
  end component ddr4_4g_1600;

END ddr_pkg;

PACKAGE BODY ddr_pkg IS   

END ddr_pkg;

