-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Connect the clocks for the tech_eth_10g.
-- Description:
--   Apply the clocks from top level down, such that they have their
--   rising_edge() aligned without any delta-delay. The reference clocks
--   are also reassigned to preserve same delta-cycle phase as for the data
--   clocks.


LIBRARY IEEE, technology_lib;
USE IEEE.std_logic_1164.ALL;
USE technology_lib.technology_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY tech_eth_10g_clocks IS
  GENERIC (
    g_technology          : NATURAL := c_tech_select_default;
    g_nof_channels        : NATURAL := 1
  );
  PORT (
    -- Input clocks
    -- . Reference
    tr_ref_clk_644    : IN  STD_LOGIC := 'X';  -- 10GBASE-R
    tr_ref_clk_312    : IN  STD_LOGIC := 'X';  -- 10GBASE-R
    tr_ref_clk_156    : IN  STD_LOGIC := 'X';  -- 10GBASE-R or XAUI
    tr_ref_rst_156    : IN  STD_LOGIC := 'X';  -- 10GBASE-R or XAUI
    
    -- . XAUI
    tx_rst_arr        : IN  STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0) := (OTHERS=>'0');
    rx_clk_arr        : IN  STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0) := (OTHERS=>'X');
    rx_rst_arr        : IN  STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0) := (OTHERS=>'0');
    
    -- Output clocks
    -- . Reference
    eth_ref_clk_644   : OUT STD_LOGIC;
    eth_ref_clk_312   : OUT STD_LOGIC;
    eth_ref_clk_156   : OUT STD_LOGIC;
    eth_ref_rst_156   : OUT STD_LOGIC;
    
    -- . Data
    eth_tx_clk_arr    : OUT STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
    eth_tx_rst_arr    : OUT STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
    
    eth_rx_clk_arr    : OUT STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
    eth_rx_rst_arr    : OUT STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0)
  );
END tech_eth_10g_clocks;


ARCHITECTURE str OF tech_eth_10g_clocks IS
BEGIN

  -- Reference
  eth_ref_clk_644 <= tr_ref_clk_644;
  eth_ref_clk_312 <= tr_ref_clk_312;
  eth_ref_clk_156 <= tr_ref_clk_156;
  eth_ref_rst_156 <= tr_ref_rst_156;
  
  gen_clocks_xaui : IF g_technology=c_tech_stratixiv GENERATE  
    -- Data
    eth_tx_clk_arr  <= (OTHERS=>tr_ref_clk_156);
    eth_tx_rst_arr  <= tx_rst_arr;
    
    eth_rx_clk_arr  <= rx_clk_arr;
    eth_rx_rst_arr  <= rx_rst_arr;
  END GENERATE;
    
  gen_clocks_10gbase_r : IF g_technology=c_tech_arria10 GENERATE
    eth_tx_clk_arr  <= (OTHERS=>tr_ref_clk_156);
    eth_tx_rst_arr  <= (OTHERS=>tr_ref_rst_156);
    
    eth_rx_clk_arr  <= (OTHERS=>tr_ref_clk_156);
    eth_rx_rst_arr  <= (OTHERS=>tr_ref_rst_156);
  END GENERATE;
  
END str;
